"""
    request 去重

    setting 启用才会去重
    请求过来不论设置不设置 filter_repeat 都会加入去重
    只有设置 filter_repeat 才会对请求做 阻拦 操作
"""
from palp import settings
from palp.filter.filter_bloom import BloomFilter
from palp.filter.filter_redis import RequestRedisFilter
from palp.filter.filter_memory import RequestMemoryFilter
from palp.exception.exception_drop import DropRequestException
from palp.filter.filter_redis_bloom import RequestRedisBloomFilter
from palp.middleware.middleware_request_base import RequestMiddleware


# 本地：基于 python set 的去重
class RequestMemoryFilterMiddleware(RequestMiddleware):
    def __init__(self):
        self.request_memory_filter = RequestMemoryFilter()

    def request_in(self, spider, request) -> None:
        if settings.REQUEST_FILTER:
            is_repeat = self.request_memory_filter.is_repeat(obj=request)

            if request.filter_repeat and is_repeat:
                raise DropRequestException(f"丢弃重复请求：{request}")


# 本地：基于 bloom 的去重
class RequestBloomFilterMiddleware(RequestMiddleware):
    def __init__(self):
        self.request_bloom_filter = BloomFilter()

    def request_in(self, spider, request) -> None:
        if settings.REQUEST_FILTER:
            is_repeat = self.request_bloom_filter.is_repeat(obj=request)

            if request.filter_repeat and is_repeat:
                raise DropRequestException(f"丢弃重复请求：{request}")


# redis：基于 redis set 的去重
class RequestRedisFilterMiddleware(RequestMiddleware):
    def __init__(self):
        self.request_redis_filter = RequestRedisFilter()

    def request_in(self, spider, request) -> None:
        if settings.REQUEST_FILTER:
            is_repeat = self.request_redis_filter.is_repeat(obj=request)

            if request.filter_repeat and is_repeat:
                raise DropRequestException(f"丢弃重复请求：{request}")


# redis：基于 redis 的 bloom 去重
class RequestRedisBloomFilterMiddleware(RequestMiddleware):
    def __init__(self):
        self.request_redis_bloom_filter = RequestRedisBloomFilter()

    def request_in(self, spider, request) -> None:
        if settings.REQUEST_FILTER:
            is_repeat = self.request_redis_bloom_filter.is_repeat(obj=request)

            if request.filter_repeat and is_repeat:
                raise DropRequestException(f"丢弃重复请求：{request}")



"""
    创建 spider 文件
"""
import re
import datetime
from pathlib import Path
from typing import Union
from palp.exception.exception_error import SpiderHasExistsError


class CreateSpider:
    def __init__(self, spider_name: str, spider_type: int = 1):
        """

        :param spider_name: spider 名字
        :param spider_type: spider 类型（1 为普通 spider，2 为分布式 spider）
        """
        self.spider_name = spider_name
        self.spider_type = spider_type
        self.path = Path('.').absolute()  # 相对路径
        self.path_template = Path(__file__).absolute().parent.parent.joinpath('template').joinpath('spider')  # 模板路径
        if self.spider_type == 2:
            self.path_template_spider = self.path_template.joinpath('distributive_spider.tmpl')
        else:
            self.path_template_spider = self.path_template.joinpath('spider.tmpl')

    def create(self):
        """
        创建 spider.py

        :return:
        """
        self.check_name()

        spider_dir = self.find_path(path=self.path)
        if not spider_dir:
            spider_dir = self.path
        if spider_dir.joinpath(self.spider_name.lower() + '.py').exists():
            raise SpiderHasExistsError(f'{self.spider_name.lower()} 已存在！')

        with open(self.path_template_spider, 'r', encoding='utf-8') as f:
            content = f.read()

        content = content.replace('${SPIDER_NAME}', self.spider_name.title())
        content = content.replace('${SPIDER_NAME_LOWER}', self.spider_name.lower())
        content = content.replace('${DATE}', str(datetime.datetime.now()))

        with open(spider_dir.joinpath(self.spider_name.lower() + '.py'), 'w', encoding='utf-8') as f:
            f.write(content)

    def check_name(self):
        """
        检查名字是否符合要求

        :return:
        """
        matcher = re.match(r'[0-9a-zA-Z_]+', self.spider_name)

        if not re.match(r'[0-9_]', self.spider_name) and matcher and len(matcher.group()) == len(self.spider_name):
            return True

        raise NameError('Spider 名字请以字母开头，并仅含有数字字母下划线！')

    def find_path(self, path: Path, deep: int = 0) -> Union[Path, None]:
        """
        寻找 spiders 文件夹的位置

        :param path:
        :param deep: 递归深度
        :return:
        """
        if deep > 3:
            return

        if path.is_dir() and path.name == 'spiders':
            return path

        for spider_dir in path.iterdir():
            if spider_dir.is_dir():
                spider_dir = self.find_path(spider_dir, deep=deep + 1)
                if spider_dir:
                    return spider_dir


if __name__ == '__main__':
    CreateSpider(spider_name='demo', spider_type=1).create()



"""
    请求检查中间件
"""
from loguru import logger
from palp import settings
from urllib.parse import urlparse
from palp.exception.exception_drop import DropRequestException
from palp.middleware.middleware_request_base import RequestMiddleware


class RequestCheckMiddleware(RequestMiddleware):
    def request_in(self, spider, request) -> None:
        # 判断使用过滤时是否开启了过滤
        if request.filter_repeat and not settings.REQUEST_FILTER:
            logger.warning(f'过滤请求请将 settings.REQUEST_FILTER 设置为 True，当前请求：{request.url}')

        # 判断域名是否在可用域名内
        domains = spider.__class__.spider_domains
        if isinstance(domains, list) and len(domains) != 0:
            domain = urlparse(request.url).netloc
            if domain not in domains:
                raise DropRequestException()



"""
    request 失败回收中间件
"""
import json
from palp import settings
from palp.middleware.middleware_request_base import RequestMiddleware


class RequestRecycleMiddleware(RequestMiddleware):
    def request_failed(self, spider, request) -> None:
        """
        回收失败的请求，为了能够直观的展示，就不使用 pickle 了

        注意：开启 REQUEST_FAILED_SAVE 才会启用

        :param spider:
        :param request:
        :return:
        """
        from palp.conn import redis_conn

        req = request.to_dict()
        if req['callback']:
            req['callback'] = req['callback'].__name__
        req['cookies'] = req['cookie_jar'].get_dict()

        del req['cookie_jar']
        del req['session']

        redis_conn.sadd(settings.REDIS_KEY_QUEUE_BAD_REQUEST, json.dumps(req, ensure_ascii=False))



"""
    通过 item 将数据传输至 pipeline
"""
import palp


class Item(palp.Item):
    def __init__(self, **kwargs):
        # 懒人方式
        for key, value in kwargs.items():
            self[key] = value

        # 一般方式
        # self.xxx = kwargs.get('xxx')



"""
    队列结构
"""
from abc import abstractmethod


class BaseSequence:

    @abstractmethod
    def put(self, obj, timeout: int = None):
        """
        添加任务

        :param timeout:
        :param obj
        :return:
        """
        pass

    @abstractmethod
    def get(self, timeout: int = None):
        """
        获取任务

        :return:
        """
        pass

    @abstractmethod
    def empty(self) -> bool:
        """
        判断队列是否为空

        :return:
        """
        pass



"""
    redis 队列
    使用了 pickle 并使用 zlib 压缩
"""
import zlib
import pickle
from palp import settings
from palp.sequence.sequence_base import BaseSequence


# 先进先出队列
class FIFOSequence(BaseSequence):
    def put(self, obj, timeout: int = None):
        """
        添加任务

        :param obj:
        :param timeout:
        :return:
        """
        from palp.conn import redis_conn

        redis_conn.rpush(settings.REDIS_KEY_QUEUE_REQUEST, zlib.compress(pickle.dumps(obj)))

    def get(self, timeout: int = None):
        """
        获取任务（这里是返回的对象）

        :return:
        """
        from palp.conn import redis_conn

        result = redis_conn.blpop(settings.REDIS_KEY_QUEUE_REQUEST, timeout=timeout)
        if result:
            return pickle.loads(zlib.decompress(result[-1]))  # 这里不需要 decode 因为是对象

    def empty(self):
        """
        判断队列是否为空

        :return:
        """
        from palp.conn import redis_conn

        return redis_conn.llen(settings.REDIS_KEY_QUEUE_REQUEST) == 0


# 后进先出队列
class LIFOSequence(FIFOSequence):
    def get(self, timeout: int = None):
        """
        获取任务
        :return:
        """
        from palp.conn import redis_conn

        result = redis_conn.brpop(settings.REDIS_KEY_QUEUE_REQUEST, timeout=timeout)
        if result:
            return pickle.loads(zlib.decompress(result[-1]))


# 优先级队列
class PrioritySequence(BaseSequence):
    def put(self, obj, timeout: int = None):
        """
        添加任务

        :param obj:
        :param timeout:
        :return:
        """
        from palp.conn import redis_conn

        redis_conn.zadd(settings.REDIS_KEY_QUEUE_REQUEST, {zlib.compress(pickle.dumps(obj)): obj.level})

    def get(self, timeout: int = None):
        """
        获取任务（这里是返回的对象）

        :return:
        """
        from palp.conn import redis_conn

        result = redis_conn.bzpopmin(settings.REDIS_KEY_QUEUE_REQUEST, timeout=timeout)
        if result:
            return pickle.loads(zlib.decompress(result[1]))

    def empty(self):
        """
        判断队列是否为空

        :return:
        """
        from palp.conn import redis_conn

        return redis_conn.zcard(settings.REDIS_KEY_QUEUE_REQUEST) == 0



"""
    本地布隆过滤器

    print(BloomFilter.is_repeat(request=RequestGet(url='https://www.baidu.com')))
    print(BloomFilter.is_repeat(request=RequestGet(url='https://www.baidu.com')))
"""
from palp import settings
from palp.network.request import Request
from pybloom_live import ScalableBloomFilter
from palp.filter.filter_base import BaseFilter, FilterLock


class BloomFilter(BaseFilter):
    def __init__(self):
        self.bloom_filter_request = ScalableBloomFilter()
        self.bloom_filter_item = ScalableBloomFilter()

    def is_repeat(self, obj, **kwargs) -> bool:
        """
        获取对应的指纹，通过 本地 的 布隆过滤器 去重

        :param obj:
        :param kwargs:
        :return:
        """
        fingerprint = self.fingerprint(obj=obj)

        if isinstance(obj, Request):
            bloom_filter = self.bloom_filter_request
        else:
            bloom_filter = self.bloom_filter_item

        if settings.STRICT_FILTER:
            with FilterLock():
                return self.judge(fingerprint, bloom_filter)
        else:
            return self.judge(fingerprint, bloom_filter)

    def judge(self, fingerprint, f) -> bool:
        """
        进行判断

        :param fingerprint:
        :param f:
        :return:
        """
        return f.add(fingerprint)



from palp.network.request import Request


class RequestGet(Request):
    def __init__(self, url, **kwargs):
        kwargs.update({'url': url, 'method': 'GET'})
        super().__init__(**kwargs)


class RequestPost(Request):
    def __init__(self, url, **kwargs):
        kwargs.update({'url': url, 'method': 'POST'})
        super().__init__(**kwargs)


class RequestOptions(Request):
    def __init__(self, url, **kwargs):
        kwargs.update({'url': url, 'method': 'OPTIONS'})
        super().__init__(**kwargs)


class RequestHead(Request):
    def __init__(self, url, **kwargs):
        kwargs.update({'url': url, 'method': 'HEAD'})
        super().__init__(**kwargs)


class RequestPatch(Request):
    def __init__(self, url, **kwargs):
        kwargs.update({'url': url, 'method': 'PATCH'})
        super().__init__(**kwargs)


class RequestDelete(Request):
    def __init__(self, url, **kwargs):
        kwargs.update({'url': url, 'method': 'DELETE'})
        super().__init__(**kwargs)


if __name__ == '__main__':
    response = RequestGet(url='https://www.baidu.com', titan_meta={'name': '张三'}).send()
    print(response.text)
    print(response.titan)
    print(response.titan_meta)



"""
    单机 spider
"""
from palp import settings
from threading import Thread
from abc import abstractmethod
from palp.spider.spider_base import BaseSpider


class Spider(BaseSpider, Thread):
    def __init__(self, thread_count=None, request_filter=False, item_filter=False):
        """

        :param thread_count: 线程数量
        :param request_filter: 开启请求过滤，不然 filter_repeat 无效
        :param item_filter: 开启 item 过滤
        """
        setattr(settings, 'SPIDER_TYPE', 1)
        super().__init__(thread_count, request_filter, item_filter)

    @abstractmethod
    def start_requests(self) -> None:
        """
        起始函数

        :return:
        """
        pass

    def parse(self, request, response) -> None:
        """
        解析

        :param request:
        :param response:
        :return:
        """
        pass

    def spider_logic(self):
        """
        spider 处理逻辑

        :return:
        """
        # 分发任务
        self.distribute_task()



"""
    spider 资源回收中间件
"""
from palp import conn
from palp.middleware.middleware_spider_base import SpiderMiddleware


class SpiderRecycleMiddleware(SpiderMiddleware):
    def spider_close(self, spider) -> None:
        """
        关闭所有创建的连接

        :param spider:
        :return:
        """
        if conn.redis_conn:
            conn.redis_conn.close()
        if conn.pg_conn:
            conn.pg_conn.engine.dispose()
        if conn.mysql_conn:
            conn.mysql_conn.engine.dispose()
        if conn.mongo_conn:
            conn.mongo_conn.close()
        if conn.kafka_conn:
            conn.kafka_conn.close()



"""
    redis 批次爬虫

    注意：
        redis_key 为 spider_name:request 且不可更改
        若使用优先级队列，则放进去的数据形式应该为 {xxx:级别}
"""
import palp
from palp import sequence
from abc import abstractmethod


class BatchRedisSpider(palp.DistributiveSpider):
    spider_key = 'baidu:request'
    spider_queue = sequence.RequestRedisFIFOSequence()  # 队列类型

    def start_batch(self):
        """
        开启一个批次

        可以从任何地方获取批次

        :return:
        """
        while True:
            if self.__class__.spider_queue.empty():
                break

            url = self.__class__.spider_queue.get(timeout=5)

            if url:
                yield palp.RequestGet(url)

    def start_requests(self) -> None:
        """
        起始函数获取批次

        :return:
        """

    @abstractmethod
    def parse(self, request, response) -> None:
        """
        解析函数

        :param request:
        :param response:
        :return:
        """
        pass



"""
    用来抛出错误
"""


# 函数不含有 yield
class NotGeneratorFunctionError(Exception):
    def __init__(self, *args):
        super(NotGeneratorFunctionError, self).__init__(*args)


# spider 已存在错误
class SpiderHasExistsError(Exception):
    def __init__(self, *args):
        super(SpiderHasExistsError, self).__init__(*args)



"""
    中间件
    SpiderMiddleware：爬虫中间件
    RequestMiddleware：请求中间件

    参数全部可直接原地修改
"""
import palp
from loguru import logger


class SpiderMiddleware(palp.SpiderMiddleware):
    def spider_start(self, spider) -> None:
        """
        spider 开始时的操作

        :param spider:
        :return:
        """
        pass

    def spider_close(self, spider) -> None:
        """
        spider 结束的操作

        :param spider:
        :return:
        """
        pass

    def spider_error(self, spider, exception_type: str, exception: str) -> None:
        """
        spider 出错时的操作

        :param spider:
        :param exception_type: 错误的类型
        :param exception: 错误的详细信息
        :return:
        """
        logger.error(exception)


class RequestMiddleware(palp.RequestMiddleware):
    def request_in(self, spider, request) -> None:
        """
        请求进入时的操作

        :param spider:
        :param request:
        :return:
        """
        pass

    def request_close(self, spider, request, response):
        """
        请求结束时的操作

        :param spider:
        :param request: 该参数可返回（用于放弃当前请求，并发起新请求）
        :param response:
        :return: [Request, None]
        """
        return

    def request_error(self, spider, request, exception_type: str, exception: str):
        """
        请求出错时的操作

        :param spider:
        :param request: 该参数可返回（用于放弃当前请求，并发起新请求）
        :param exception_type: 错误的类型
        :param exception: 错误的详细信息
        :return: [Request, None]
        """
        logger.error(exception)

        return

    def request_failed(self, spider, request) -> None:
        """
        超过最大重试次数时的操作

        :param spider:
        :param request:
        :return:
        """
        logger.warning(f"失败的请求：{request}")



from palp.network.request import Request
from palp.network.response import Response
from palp.network.request_method import RequestGet, RequestPost, RequestDelete, RequestOptions, RequestHead, \
    RequestPatch



from setuptools import setup, find_packages  # 这个包没有的可以pip一下

file_path = './README.md'

setup(
    name="palp",  # 这里是pip项目发布的名称
    version="0.0.76",  # 版本号，数值大的会优先被pip
    keywords=["palp"],  # 关键字
    description="一个 爬虫框架",  # 描述
    long_description=open(file_path, 'r', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    license="MIT Licence",  # 许可证

    url="https://github.com/Leviathangk/palp",  # 项目相关文件地址，一般是github项目地址即可
    author="郭一会儿",  # 作者
    author_email="1015295213@qq.com",

    packages=find_packages(),
    include_package_data=True,
    platforms="any",
    install_requires=['zmail', 'requests', 'quickdb', 'pybloom-live', 'parsel', 'loguru'],

    # 使得可以命令行访问
    entry_points={
        'console_scripts': [
            'palp = palp.tool.shell:main'
        ]
    }
)



"""
    过滤器
"""
import hashlib
import threading
from typing import Union
from abc import abstractmethod
from urllib.parse import urlencode
from palp.item.item_base import BaseItem
from palp.network.request import Request


class BaseFilter:
    @abstractmethod
    def is_repeat(self, obj: Union[Request, BaseItem], **kwargs) -> bool:
        """
        判断是否重复，必须实现不存在则添加的方法

        :param obj: request 对象或 item 对象
        :return:
        """
        pass

    @abstractmethod
    def judge(self, fingerprint, f) -> bool:
        """
        进行判断

        :param fingerprint: 指纹
        :param f: 判断条件或方法之类
        :return:
        """
        pass

    @staticmethod
    def fingerprint(obj: Union[Request, BaseItem]) -> str:
        """
        获取能够代表唯一的字符串

        :param obj: request 对象或 item 对象
        :return:
        """
        if isinstance(obj, Request):
            if obj.method == 'GET':
                filter_str = urlencode(obj.params or {})
            elif obj.method == 'POST':
                filter_str = urlencode(obj.params or {}) + urlencode(obj.data or {}) + urlencode(obj.json or {})
            else:
                filter_str = ''

            filter_str = obj.method + obj.url + filter_str
        else:
            filter_str = obj.to_json()

        return hashlib.md5(filter_str.encode()).hexdigest()


# filter 专用的锁
class FilterLock:
    lock = threading.RLock()

    def __init__(self, timout: int = 5):
        """
        锁的时间

        :param timout:
        """
        self.timeout = timout

    def __enter__(self):
        self.__class__.lock.acquire(timeout=self.timeout)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__class__.lock.release()



from palp.sequence.sequence_redis_request import FIFOSequence as RequestRedisFIFOSequence
from palp.sequence.sequence_redis_request import LIFOSequence as RequestRedisLIFOSequence
from palp.sequence.sequence_redis_request import PrioritySequence as RequestRedisPrioritySequence
